/* (c) 2009-2010 Botik-Technologies Ltd. http://www.botik.ru
   Distributed under the terms of the BSD open-source license.
   See license.txt for more information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include "md5.h"

#define APPNAME "BotikKeyC"
#define VER "1.0"
#define MSGSIZE 768
#define SOCK_TIMEO 5000
#define MAXGATES 10
#define IPLEN 16
#define BKPORT 12040
#define LHOSTX 0x7f000001

#define add_gateway(addr, gw) strncpy(gw,addr,IPLEN)

#ifdef __APPLE__
#define _MAC
int  kchain_item_exists(ucharp dpassword);
void kchain_item_add(ucharp dpassword);
void kchain_item_modify(ucharp dpassword);
#else
#define kchain_item_exists(d) 0
#define kchain_item_add(d)
#define kchain_item_modify(d)
#endif /* __APPLE__ */

enum Color {Black, Red, Green, Yellow, Blue, Magenta, Cyan, White};

int parse_args(int, char**, char*, int, ucharp, int);
int main_loop(char *server, ucharp dpassword);
void get_password(char *to, unsigned int length, int fd, int echo);
int read_line(char *path, char *pwd, int count);
void get_response(char *challenge, char *dpassword, char *result);
void trim(char *str);
int connect_to(const char *nodeName, int port);
void print_usage(void);
int create_socket(int nonblock);
int invalid_option (char opt);
int local_server(void);
void exit_app(int code);
void signal_handler(int sig);
char *new_string (int length);
ucharp new_ustring (int length);
void myprintf(int clr, const char *format, ...);
void query_password(char *to, int size);
int init_library(void);
int myrand(void);
int sock_readable(int sock);
int sock_writable(int sock);
int find_gateways(char gws[][IPLEN], int skip);
int other_clients(void);
void write_log(char *str);
void color_on(int color);
void color_off(void);
int color_capable(void);
void init_colors(void);

extern int verbose;
extern int quiet;
extern int coloron;
extern char *logfile;

#ifndef _WIN32
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <net/if.h>
#include <fcntl.h>
#include <netdb.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/times.h>
#include <sys/time.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <termios.h>
#include <stdarg.h>
#include <poll.h>
#ifdef _MAC
#include <pwd.h>
#include <Security/Security.h>
#include <CoreFoundation/CoreFoundation.h>
#endif

#define TERMIO struct termios
#define get_last_neterr() errno
#define get_last_error() errno
#define unload_netlib()
#define close_socket(s) close(s)
#define ioctlsocket ioctl
#define set_title(x)
#define WOULDBLOCK EWOULDBLOCK
#define INPROGRESS EINPROGRESS 
#define do_sleep(ms) usleep(ms*1000)
#define _O_RDONLY O_RDONLY
#define _O_APPEND O_APPEND
#define _O_CREAT  O_CREAT
#define _O_WRONLY O_WRONLY
#define _S_IWRITE S_IWUSR|S_IRUSR
#define _vsnprintf vsnprintf
#define _snprintf snprintf
#define _open   open
#define _read   read
#define _write  write
#define _close  close
#define _stat   stat
#define _unlink unlink
#define _strdup strdup

typedef struct hostent* PHOSTENT;
typedef struct sockaddr saddr;
typedef struct stat stats_t;
clock_t get_tick_count(void);
int resolve_host(struct sockaddr_in *destaddr, const char *host);
void print_error(char* msg, int code);

#else /* Windows-specific */

#include <winsock2.h>
#include <iphlpapi.h>
#include <time.h>
#include <fcntl.h>
#include <io.h>
#include <sys/stat.h>
#include <psapi.h>
#include <windows.h>
#include <conio.h>
#define get_last_neterr() WSAGetLastError()
#define unload_netlib() WSACleanup()
#define get_last_error() GetLastError()
#define set_title(x) SetConsoleTitle(x)
#define do_sleep(ms) Sleep(ms)
#define get_tick_count() GetTickCount()
#define close_socket(s) closesocket(s)
#define WOULDBLOCK WSAEWOULDBLOCK
#define INPROGRESS WSAEINPROGRESS

typedef SOCKADDR saddr;
typedef struct _stat stats_t;
int resolve_host(struct sockaddr_in *destaddr, const char *host);
void print_error(char* msg, unsigned long code);

#endif /* !_WIN32 */
