/*
  (c) 2009-2010 Botik-Technologies Ltd. http://www.botik.ru
  Distributed under the terms of the BSD open-source license.
  See license.txt for more information.
  
  generic and cross-platform stuff for BotikKey C client
 */

#include "bkey.h"

/*******************************************/
void myprintf(int clr, const char *format, ...) {
    char *str, *chr;
    va_list args;
    time_t now;
    struct tm *current;
    int len=MSGSIZE, p;

    va_start(args,format);
    str = new_string(len);
    now = time(0);
    current=localtime(&now);
    p=_snprintf(str,len,"[%2d:%2d:%2d]",current->tm_hour,current->tm_min,current->tm_sec);
    while ((chr = strchr(str,' ')) != NULL) chr[0] = '0'; /* replace spaces with zeroes */
    p+=_snprintf(str+p,len-p," ");
    _vsnprintf(str+p,len-p,format,args);
    va_end(args);
    if (!quiet) {
        color_on(clr);
        printf("%s",str);
        color_off();
    }
    write_log(str);
    free(str);
}
/*******************************************/
void write_log(char *str) {
    int fd, r;
    stats_t st;

    if (!logfile) return;
    fd = _open(logfile,_O_CREAT|_O_APPEND|_O_WRONLY,_S_IWRITE);
    if (fd < 0) return;
    r = _write(fd,str,strlen(str));
    if (r < 0) return;
    r = _close(fd);
    if (r < 0) return;
    r = _stat(logfile, &st);
    if (r < 0) return;
    /* delete log file if its size exceeds 1 MB */
    if (st.st_size > 1024*1024) {
        _unlink(logfile);
    }
}
/*******************************************/
ucharp new_ustring (int length) {
    ucharp addr = (ucharp)malloc(length*sizeof(uchar));
    if (addr==NULL) {
        myprintf(Red,"Error allocating %d bytes of memory.\n", length*sizeof(uchar));
        exit_app(EXIT_FAILURE);
    }
    return addr;
}
/*******************************************/
char *new_string (int length) {
    char *addr = (char*)malloc(length*sizeof(char));
    if (addr==NULL) {
        myprintf(Red,"Error allocating %d bytes of memory.\n", length*sizeof(char));
        exit_app(EXIT_FAILURE);
    }
    return addr;
}

#ifndef _WIN32
/*******************************************/
void color_on(int color) {
    char colorFormat[] = "\033[0;30m";

    if (!coloron || !color_capable()) return;
    colorFormat[2] += color >>3; /* bold or not */
    colorFormat[5] += color & 7; /* removing brightness */
    fputs(colorFormat,stdout);
}
/*******************************************/
void color_off() {
    if (!coloron || !color_capable()) return;
    printf("\033[0;39m");
}
/*******************************************/
int color_capable() {
    return isatty(1) || getenv("TTY_HAS_COLORS");
}
/*******************************************/
void init_colors() {
    /*
        no need to initialize colors in UNIX
    */
}
/******************************************
   checks if other BKey clients are running,
   possibly this will not work properly in OSes other than Linux and Mac OS X,
   because netstat's output differs in different OSes
*/
int other_clients() {
    return 0;
/*
    FILE *in;
    char tmp[128], *chr;
    int line=0, r=0, nf;
    char data[9][128];

    in=popen("netstat -nt", "r");
    while(fgets(tmp,sizeof(tmp),in)!=NULL) {
        if (++line < 3) continue;
        nf = sscanf(tmp,"%s %s %s %s %s %s %s %s %s",data[0],data[1],data[2],data[3],data[4],data[5],data[6],data[7],data[8]);
        if (nf != 6) continue;
#ifdef _MAC
        chr=strrchr(data[4],'.');
#else
        chr=strchr(data[4],':');
#endif
        if (atoi(chr+1)==BKPORT && strncmp(data[5],"ESTABLISHED",11)==0) {
            r=1;
            break;
        }
    }
    pclose(in);
    return r;
*/
}
/*******************************************/
int sock_readable(int sock) {
    struct pollfd *fds;
    int r;

    fds=(struct pollfd*)malloc(sizeof(struct pollfd));
    if (fds==NULL) {
        myprintf(Red,"Error allocating memory for fds.\n");
        exit_app(EXIT_FAILURE);
    }
    fds->fd=sock;
    fds->events=POLLIN;
    fds->revents=0;
    r=poll(fds,1,SOCK_TIMEO);
    free(fds);
    return (r > 0);
}
/*******************************************/
int sock_writable(int sock) {
    struct pollfd *fds;
    int r;

    fds=(struct pollfd*)malloc(sizeof(struct pollfd));
    if (fds==NULL) {
        myprintf(Red,"Error allocating memory for fds.\n");
        exit_app(EXIT_FAILURE);
    }
    fds->fd=sock;
    fds->events=POLLOUT;
    fds->revents=0;
    r=poll(fds,1,SOCK_TIMEO);
    free(fds);
    return (r > 0);
}
/*******************************************/
int find_gateways(char gws[][IPLEN], int skip) {
    FILE *in;
    char tmp[128];
    int line=0, n=0;
    char data[8][IPLEN];

    tmp[0]='\0';
    in=popen("netstat -rn", "r");
    while(fgets(tmp,sizeof(tmp),in)!=NULL) {
        line++;
        if (line < 3) continue; /* skip first 2 lines */
        sscanf(tmp,"%s %s %s %s %s %s %s %s",data[0],data[1],data[2],data[3],data[4],data[5],data[6],data[7]);
        if (strncmp(data[0],"0.0.0.0",7) == 0 || strncmp(data[0],"default",7) == 0) {
            if (n == MAXGATES - skip) break;
            add_gateway(data[1],gws[n++]);
        }
    }
    pclose(in);
    return n;
}
/*******************************************/
void get_password(char *to, unsigned int length, int fd, int echo) {
    char *pos=to,*end=to+length, tmp;

    for (;;) {
        if (read(fd,&tmp,1) != 1)
            break;
        if (tmp == '\b' || (int) tmp == 127) {
            if (pos != to) {
                if (echo) {
                    fputs("\b \b",stdout);
                    fflush(stdout);
                }
                pos--;
                continue;
            }
        }
        if (tmp == '\n' || tmp == '\r' || (int)tmp == 3)
            break;
        if (iscntrl(tmp) || pos == end)
            continue;
        if (echo) {
            fputc('*',stdout);
            fflush(stdout);
        }
        *(pos++) = tmp;
    }
    while (pos != to && isspace(pos[-1]) == ' ')
        pos--;
    *pos=0;
    return;
}
/*******************************************/
void query_password(char *buff, int size) {
    TERMIO org,tmp;
    int atty;

    color_on(Yellow);
    if (isatty(fileno(stdout))) {
        fputs("Enter BotikKey password: ",stdout);
        fflush(stdout);
    }
    tcgetattr(fileno(stdin), &org);
    tmp = org;
    tmp.c_lflag &= ~(ECHO | ISIG | ICANON);
    tmp.c_cc[VMIN] = 1;
    tmp.c_cc[VTIME] = 0;
    tcsetattr(fileno(stdin), TCSADRAIN, &tmp);
    atty = isatty(fileno(stdout));
    get_password(buff, size-1, fileno(stdin), atty);
    tcsetattr(fileno(stdin), TCSADRAIN, &org);
    if (isatty(fileno(stdout)))
        fputc('\n',stdout);
    color_off();
}
/*******************************************/
void print_error(char* msg, int code) {
    myprintf(Red,"%s failed with error: %s\n",msg,strerror(code));
}
/*******************************************/
int init_library() {return 0;}
/*******************************************/
int myrand() {
    return (int)((rand()/65536)%32768);
}
/*******************************************/
clock_t get_tick_count() {
    struct tms tm;
    return times(&tm);
}

#ifdef _MAC
/*******************************************/
int kchain_item_exists(ucharp dpassword) {
    UInt32 dlen=0, namelen;
    struct passwd *passwd;
    char *username=new_string(30);
    OSStatus st;
    void *dpass;

    passwd = getpwuid(getuid());
    namelen = strlen(passwd->pw_name);
    strncpy(username,passwd->pw_name,namelen);
    /* query MacOS Keychain for BotikKey service password */
    st = SecKeychainFindGenericPassword (
        NULL,
        strlen(APPNAME),
        APPNAME,
        namelen,
        username,
        &dlen,
        &dpass,
        NULL);
    if (st == 0)
        strncpy((char*)dpassword,(char*)dpass,dlen);
    return (st == 0);
}
/*******************************************/
void kchain_item_add(ucharp dpassword) {
    int namelen, dlen=2*MD5_HASHBYTES+1;
    struct passwd *passwd;
    char *username=new_string(30);
    const char *errstr;
    CFStringRef strdesc;
    OSStatus st;

    passwd = getpwuid(getuid());
    namelen = strlen(passwd->pw_name);
    strncpy(username,passwd->pw_name,namelen);
    st = SecKeychainAddGenericPassword (
        NULL,
        strlen(APPNAME),
        APPNAME,
        namelen,
        username,
        dlen,
        (void*)dpassword,
        NULL);
    if (st) {
        strdesc = (CFStringRef)SecCopyErrorMessageString(st,NULL);
        errstr = CFStringGetCStringPtr(strdesc, kCFStringEncodingMacRoman);
        myprintf(Red,"Keychain: %s\n",errstr);
    }
}
/*******************************************/
void kchain_item_modify(ucharp dpassword) {
    int dlen=2*MD5_HASHBYTES+1, namelen;
    struct passwd *passwd;
    char *username=new_string(30);
    SecKeychainItemRef item;
    OSStatus st;
    const char *errstr;
    CFStringRef strdesc;

    passwd = getpwuid(getuid());
    namelen = strlen(passwd->pw_name);
    strncpy(username,passwd->pw_name,namelen);
    st = SecKeychainFindGenericPassword (
        NULL,
        strlen(APPNAME),
        APPNAME,
        namelen,
        username,
        NULL,
        NULL,
        &item);
    if (st==0) {
        /* item found */
        st = SecKeychainItemModifyContent(
            item,
            NULL,
            dlen,
            (void*)dpassword);
        if (st != 0) {
            /* error when modifying it */
            strdesc = (CFStringRef)SecCopyErrorMessageString(st,NULL);
            errstr = CFStringGetCStringPtr(strdesc, kCFStringEncodingMacRoman);
            myprintf(Red,"Keychain: cannot reset password: %s\n",errstr);
        }
    } else if (st == errSecItemNotFound) {
        /* make new item */
        kchain_item_add(dpassword);
    } else {
        strdesc = (CFStringRef)SecCopyErrorMessageString(st,NULL);
        errstr = CFStringGetCStringPtr(strdesc, kCFStringEncodingMacRoman);
        myprintf(Red,"Keychain: cannot reset password. %s\n",errstr);
    }
}
#endif /* _MAC  */

#else /* Windows-specific code */

HANDLE hConsoleHandle;
WORD originalColors;
/*******************************************/
void set_color(int color) {
  switch (color) {
  case 0:    /* Black on Gray */
    SetConsoleTextAttribute(hConsoleHandle,
						BACKGROUND_INTENSITY | BACKGROUND_INTENSITY);
    break;
  case 1:    /* Red on Black */
    SetConsoleTextAttribute(hConsoleHandle,
                         FOREGROUND_INTENSITY | FOREGROUND_RED);
    break;
  case 2:    /* Green on Black */
    SetConsoleTextAttribute(hConsoleHandle,
						FOREGROUND_INTENSITY | FOREGROUND_GREEN);
    break;
  case 3:    /* Yellow on Black */
    SetConsoleTextAttribute(hConsoleHandle,
                      FOREGROUND_INTENSITY | FOREGROUND_RED |
                      FOREGROUND_GREEN);
    break;
  case 4:    /* Blue on Black */
    SetConsoleTextAttribute(hConsoleHandle,
                      FOREGROUND_INTENSITY | FOREGROUND_BLUE);
    break;
  case 5:    /* Magenta on Black */
    SetConsoleTextAttribute(hConsoleHandle,
                     FOREGROUND_INTENSITY | FOREGROUND_RED |
                     FOREGROUND_BLUE);
    break;
  case 6:    /* Cyan on Black */
    SetConsoleTextAttribute(hConsoleHandle,
                    FOREGROUND_INTENSITY | FOREGROUND_GREEN |
                    FOREGROUND_BLUE);
    break;
  default:    /* White on Black */
    SetConsoleTextAttribute(hConsoleHandle,
                FOREGROUND_INTENSITY | FOREGROUND_RED |
                FOREGROUND_GREEN | FOREGROUND_BLUE);
    break;
  }
}
/*******************************************/
int color_capable() {
    HANDLE oHandle;
    oHandle = GetStdHandle(STD_OUTPUT_HANDLE);
    return (oHandle!=NULL && oHandle!=INVALID_HANDLE_VALUE);
}
/*******************************************/
void color_on(int color) {
    if (!coloron || !color_capable()) return;
    set_color(color);
}
/*******************************************/
void color_off() {
    if (!coloron || !color_capable()) return;
    SetConsoleTextAttribute(hConsoleHandle,originalColors);
}
/*******************************************/
void init_colors() {
    CONSOLE_SCREEN_BUFFER_INFO *ConsoleInfo;
    if (!coloron || !color_capable()) return;
	hConsoleHandle = GetStdHandle(STD_OUTPUT_HANDLE);
    ConsoleInfo = (CONSOLE_SCREEN_BUFFER_INFO*)malloc(sizeof(CONSOLE_SCREEN_BUFFER_INFO));
    if (ConsoleInfo==NULL) {
        myprintf(Red,"Error allocating memory for ConsoleInfo.\n");
        exit_app(EXIT_FAILURE);
    }
	GetConsoleScreenBufferInfo(hConsoleHandle, ConsoleInfo);
	originalColors = ConsoleInfo->wAttributes;
}
/******************************************
   puts process basename into fname,
   returns 0 if error, string length otherwise
*/
int get_process_name(DWORD processID, char *fname) {
    HANDLE hProcess;
    int r;

    hProcess = OpenProcess(PROCESS_QUERY_INFORMATION |
                            PROCESS_VM_READ,
                            FALSE, processID);
    if (hProcess == NULL) return 0;
    r=GetModuleBaseName(hProcess, NULL, fname, MAX_PATH);
    CloseHandle(hProcess);
    return r;
}
/*******************************************/
int process_exists(char *name) {
    DWORD aProcesses[1024], nBytes, nProcesses, i;
    int r=0;
    char *fname;

    /* Get the list of process identifiers. */
    if (!EnumProcesses(aProcesses, sizeof(aProcesses), &nBytes)) return 0;

    /* Calculate how many process identifiers were returned. */
    nProcesses = nBytes/sizeof(DWORD);

    /* Print the name of the modules for each process. */
    fname = new_string(MAX_PATH);
    for (i = 0; i < nProcesses; i++) {
        r=get_process_name(aProcesses[i], fname);
        if (r == 0) continue;
        r=_strnicmp(fname,name,strlen(name));
        if (r == 0) {
            r=1; /* process found */
            break;
        } else r=0;
    }
    free(fname);
    return r;
}
/******************************************
   checks if other BKey clients are running
   (currently detects only BKPro)
*/
int other_clients() {
    return process_exists("bkpsvc32.exe");
}
/*******************************************/
int find_gateways(char gws[][IPLEN], int skip) {
    PIP_ADAPTER_INFO pAdapterInfo;
    PIP_ADAPTER_INFO pAdapter = NULL;
    DWORD dwRetVal = 0;
    ULONG ulOutBufLen = sizeof(IP_ADAPTER_INFO);
    int n=0;
    char *addrstr;

    pAdapterInfo = (IP_ADAPTER_INFO*)malloc(sizeof(IP_ADAPTER_INFO));
    if (pAdapterInfo == NULL) {
        myprintf(Red,"Error allocating memory for pAdapterInfo.\n");
        exit_app(EXIT_FAILURE);
    }
    
    /* Make an initial call to GetAdaptersInfo to get
       the necessary size into the ulOutBufLen variable
    */
    if (GetAdaptersInfo(pAdapterInfo, &ulOutBufLen) == ERROR_BUFFER_OVERFLOW) {
        free(pAdapterInfo);
        pAdapterInfo = (IP_ADAPTER_INFO*)malloc(ulOutBufLen);
        if (pAdapterInfo == NULL) {
            myprintf(Red,"Error allocating memory for pAdapterInfo.\n");
            exit_app(EXIT_FAILURE);
        }
    }
    if ((dwRetVal = GetAdaptersInfo(pAdapterInfo, &ulOutBufLen)) == NO_ERROR) {
        pAdapter = pAdapterInfo;
        while (pAdapter) {
            if (pAdapter->Type == MIB_IF_TYPE_ETHERNET) {
                /* found Ethernet adapter, now remember its gateway */
                addrstr = pAdapter->GatewayList.IpAddress.String;
                if (strncmp(addrstr,"0.0.0.0",7) > 0) {
                    if (n == MAXGATES - skip) break;
                    add_gateway(addrstr,gws[n++]);
                }
            }
            pAdapter = pAdapter->Next;
        }
    } else {
        print_error("GetAdaptersInfo",dwRetVal);

    }
    if (pAdapterInfo) free(pAdapterInfo);
    return n;
}
/*******************************************/
int sock_readable(int sock) {
    fd_set readFds;
    struct timeval timeout;

    FD_ZERO(&readFds);
    FD_SET((unsigned int)sock, &readFds);
    timeout.tv_sec = SOCK_TIMEO/1000;
    timeout.tv_usec = (SOCK_TIMEO%1000)*1000;
    return select(1, &readFds, NULL, NULL, &timeout) > 0;
}
/*******************************************/
int sock_writable(int sock) {
    fd_set writeFds;
    struct timeval timeout;

    FD_ZERO(&writeFds);
    FD_SET((unsigned int)sock, &writeFds);
    timeout.tv_sec = SOCK_TIMEO/1000;
    timeout.tv_usec = (SOCK_TIMEO%1000)*1000;
    return select(1, NULL, &writeFds, NULL, &timeout) > 0;
}
/*******************************************/
void query_password(char *to, int size) {
    char *pos=to,*end=to+size-1,tmp;

    color_on(Yellow);
    _cputs("Enter BotikKey password: ");
    for (;;) {
        tmp=(char)_getch();
        if (tmp == '\b' || (int)tmp == 127) {
            if (pos != to) {
                _cputs("\b \b");
                pos--;
                continue;
            }
        }
        if (tmp == '\n' || tmp == '\r' || tmp == 3)
            break;
        if (iscntrl(tmp) || pos == end)
            continue;
        _cputs("*");
        *(pos++) = tmp;
    }
    while (pos != to && isspace(pos[-1]) == ' ')
        pos--;
    *pos=0;
    _cputs("\n");
    color_off();
}
/*******************************************/
void print_error(char* msg, unsigned long code) {
    char* msgBuf, *resBuf;
    int len=MSGSIZE;

    msgBuf = new_string(len);
    resBuf = new_string(len);
    msgBuf[0]='\0';
    resBuf[0]='\0';
    /* Retrieve the system error message for the given code */
    FormatMessage (
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        code,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        msgBuf,
        len,
        NULL);

    CharToOemBuff(msgBuf,resBuf,len);
    myprintf(Red,"%s failed with error %d: %s", msg, code, resBuf);
    free(msgBuf);
    free(resBuf);
}
/*******************************************/
int init_library() {
    WSADATA wsaData;
    return (WSAStartup(MAKEWORD(2,2),&wsaData) != 0);
}
/*******************************************/
int myrand() {
    return (int)rand();
}
#endif /* _WIN32 */
